__metaclass__ = type

import httplib
import logging
import os
from signal import SIGKILL
import simplejson
import sys
from wsgiref.simple_server import make_server
from wsgiref.util import shift_path_info
from grackle.store import (
    MemoryStore,
    )


class GrackleService:
    """A request handler that forwards to server.store."""

    def __init__(self, store):
        self.store = store
        self.logger = logging.getLogger('http')

    def __call__(self, environ, start_response):
        self.environ = environ
        self.start_response = start_response
        self.method = environ['REQUEST_METHOD']
        self.host_port = shift_path_info(environ)
        self.application = shift_path_info(environ)
        self.path = environ['PATH_INFO'].split('/')[1:]
        self.query_string = environ['QUERY_STRING']
        return self.handle_request()

    def handle_request(self):
        if self.method == 'PUT':
            return self.do_PUT()
        if self.method == 'POST':
            return self.do_POST()
        if self.method == 'GET':
            return self.do_GET()

    def send_response(self, code, headers={}, reason=None):
        if reason is None:
            reason = httplib.responses[code]
        response_code = '%s %s' % (code, reason)
        response_headers = {'content-type': 'application/json'}
        response_headers.update(headers.items())
        self.start_response(response_code, response_headers.items())

    def do_PUT(self):
        """Create an archive or message on PUT."""
        if len(self.path) == 1:
            # This expected path is /archive/archive_id.
            try:
                self.store.put_archive(self.path[0])
                self.send_response(httplib.CREATED)
            except Exception, error:
                self.send_response(
                    httplib.BAD_REQUEST, reason=error.__doc__)
        elif len(self.path) == 2:
            # This expected path is /archive/archive_id/message_id.
            try:
                put_input = self.environ['wsgi.input']
                message = put_input.read(int(self.environ['CONTENT_LENGTH']))
                self.store.put_message(self.path[0], self.path[1], message)
                self.send_response(httplib.CREATED)
            except:
                self.send_response(httplib.BAD_REQUEST)
        return []

    def do_POST(self):
        """Change a message on POST."""
        if len(self.path) == 2:
            # This expected path is /archive/archive_id/message_id.
            try:
                # This expected path is /archive/archive_id/message_id.
                response = self.store.hide_message(
                    self.path[0], self.path[1], self.query_string)
                self.send_response(httplib.OK)
                return [simplejson.dumps(response)]
            except:
                self.send_response(httplib.BAD_REQUEST)
                return []

    def do_GET(self):
        """Retrieve a list of messages on GET."""
        try:
            response = self.store.get_messages(
                self.path[0], self.query_string)
            self.send_response(httplib.OK)
            return [simplejson.dumps(response)]
        except Exception, error:
            self.send_response(httplib.BAD_REQUEST, reason=error.__doc__)
            return []

    def log_message(self, format, *args):
        """Override log_message to use standard Python logging."""
        message = "%s - - [%s] %s\n" % (
            self.address_string(), self.log_date_time_string(), format % args)
        self.logger.info(message)


class ForkedFakeService:
    """A Grackle service fake, as a ContextManager."""

    def __init__(self, port, message_archives=None, write_logs=False):
        """Constructor.

        :param port: The tcp port to use.
        :param message_archives: A dict of lists of dicts representing
            archives of messages. The outer dict represents the archive,
            the list represents the list of messages for that archive.
        :param write_logs: If true, log messages will be written to stdout.
        """
        self.pid = None
        self.port = port
        if message_archives is None:
            self.message_archives = {}
        else:
            self.message_archives = message_archives
        self.read_end, self.write_end = os.pipe()
        self.write_logs = write_logs

    @staticmethod
    def from_client(client, message_archives=None):
        """Instantiate a ForkedFakeService from the client.

        :param port: The client to provide service for.
        :param message_archives: A dict of lists of dicts representing
            archives of messages. The outer dict represents the archive,
            the list represents the list of messages for that archive.
        """
        return ForkedFakeService(client.port, message_archives)

    def is_ready(self):
        """Tell the parent process that the server is ready for writes."""
        os.write(self.write_end, 'asdf')

    def __enter__(self):
        """Run the service.

        Fork and start a server in the child.  Return when the server is ready
        for use."""
        pid = os.fork()
        if pid == 0:
            self.start_server()
        self.pid = pid
        os.read(self.read_end, 1)
        return

    def start_server(self):
        """Start the HTTP server."""
        app = GrackleService(MemoryStore(self.message_archives))
        service = make_server('', self.port, app)
        self.is_ready()
        if self.write_logs:
            logging.basicConfig(
                stream=sys.stderr, level=logging.INFO)
        service.serve_forever()

    def __exit__(self, exc_type, exc_val, traceback):
        os.kill(self.pid, SIGKILL)


def application(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return "Hello World"


if __name__ == '__main__':
    app = GrackleService(MemoryStore({}))
    service = make_server('', 8787, app)
    service.serve_forever()
